#include <fstream>
#include <sstream>
#include <vector>
#include <iostream>
#include "scheduler.h"

scheduler::scheduler()
{
    cars.clear();
    collisionPoints.clear();
    criticalParts.clear();
    isWorking = true;

    collisionPoints.push_back(collisionPoint(1,CarColor::Blue, 285, CarColor::Yellow, 165));
    collisionPoints.push_back(collisionPoint(2,CarColor::Blue, 345, CarColor::Yellow, 105));

    collisionPoints.push_back(collisionPoint(3,CarColor::Yellow, 75, CarColor::Black, 195));
    collisionPoints.push_back(collisionPoint(4,CarColor::Yellow, 15, CarColor::Black, 255));

    collisionPoints.push_back(collisionPoint(5,CarColor::Black, 285, CarColor::Green, 165));
    collisionPoints.push_back(collisionPoint(6,CarColor::Black, 345, CarColor::Green, 105));

    collisionPoints.push_back(collisionPoint(7,CarColor::Green, 75, CarColor::Red, 195));
    collisionPoints.push_back(collisionPoint(8,CarColor::Green, 15, CarColor::Red, 255));

    criticalParts.push_back(criticalPart(CarColor::Blue, 285, Direction::Positive, 345, Direction::Negative, 
        CarColor::Yellow, 165, Direction::Negative, 105, Direction::Positive));
    criticalParts.push_back(criticalPart(CarColor::Yellow, 75, Direction::Negative, 15, Direction::Positive, 
        CarColor::Black, 195,Direction::Positive, 255, Direction::Negative));
    criticalParts.push_back(criticalPart(CarColor::Black, 285, Direction::Positive, 345, Direction::Negative, 
        CarColor::Green, 165, Direction::Negative, 105,Direction::Positive));
    criticalParts.push_back(criticalPart(CarColor::Green, 75, Direction::Negative, 15, Direction::Positive, 
        CarColor::Red, 195,Direction::Positive, 255, Direction::Negative));
}

scheduler::~scheduler()
{
}

CarColor scheduler::mapColor(std::string color)
{
    if (color == "piros") return CarColor::Red;
    if (color == "kek") return CarColor::Blue;
    if (color == "fekete") return CarColor::Black;
    if (color == "zold") return CarColor::Green;
    return CarColor::Yellow;
}

std::string scheduler::colorString(CarColor color)
{
    if (color == CarColor::Black) return "fekete";
    if (color == CarColor::Blue) return "kek";
    if (color == CarColor::Red) return "piros";
    if (color == CarColor::Green) return "zold";
    return "sarga";
}

Direction scheduler::mapDirection(std::string direction)
{
    if (direction == "+") return Direction::Positive;
    return Direction::Negative;
}

void scheduler::readInput()
{
    std::ifstream infile("kisautok.txt");
    std::string line;
    std::vector<std::string> splitted;

    CarColor color;
    int velocity;
    int pos;
    Direction dir;
    int startTime;

    while (std::getline(infile, line))
    {  
        splitted.clear();
        std::istringstream iss(line);
        for (std::string s; iss >> s; )
        {
            splitted.push_back(s);
        }

        color = mapColor(splitted[0]);
        velocity = std::stoi(splitted[1]);
        pos = std::stoi(splitted[2]);
        dir = mapDirection(splitted[3]);
        startTime = std::stoi(splitted[4]);

        cars.push_back(car(color, velocity, pos, dir, startTime));
    }
}

car& scheduler::getCarByColor(CarColor color)
{
    for (std::list<car>::iterator it = cars.begin(); it != cars.end(); ++it)
    {
        CarColor tempColor = it->getColor();
        if (tempColor == color) return *it;
    }
    
    car nullcar = car(CarColor::WrongColor, 0, 0, Direction::WrongDir, 0);
    return nullcar;
}

void scheduler::simulateOneStep(int currentTime)
{
    if (isWorking)
    {
        //Refresh the waiting flags
        refreshWaitingParameters();

        //Request step from cars
        carStep(currentTime);

        //Update arrive list
        updateArrives(currentTime);

        //Check collisions
        checkCollision();
    }
}

bool scheduler::isPosBetweenTwoPos(int check_pos, int pos1, int pos2)
{
    if (pos1 < pos2)
    {
        if (check_pos > pos1 && check_pos < pos2) return true;
    }
    else
    {
        if (check_pos < pos1 && check_pos > pos2) return true;
    }
    return false;
}

void scheduler::checkOnCarInOneCriticalPosition(car& Car1, car& Car2, int car1_pos1, int car2_pos1, int car2_pos2, 
    Direction car1_criticalDir1, Direction car2_criticalDir2)
{
    //Check if Car1 is in the critical position with critical direction (it would like to go to the critical part)
    if (Car1.getCurrentPos() == car1_pos1 && Car1.getDirection() == car1_criticalDir1)
    {
        //Check if the Car2 is in the critical part
        if (isPosBetweenTwoPos(Car2.getCurrentPos(), car2_pos1, car2_pos2))
        {
            //If the direction is same, Car1 shall wait
            if (Car1.getDirection() == Car2.getDirection()) Car1.setWaiting(true);
            //If the direction is different, Car1 shall wait until the distance from the Car2 is at least 30 deg 
            else if (abs(Car2.getCurrentPos() - car2_pos1) < 30) Car1.setWaiting(true);
        }
        //Check if Car1 is in the critical position and Car2 is in another critical position and they would like to go to the same critical part
        else if (Car2.getCurrentPos() == car2_pos2 && Car2.getDirection() == car2_criticalDir2 && Car2.getWaiting() == false) Car1.setWaiting(true);
    }
}

void scheduler::refreshWaitingParameters()
{
    for (std::list<car>::iterator it = cars.begin(); it != cars.end(); ++it)
    {
        it->setWaiting(false);
    }

    for each (criticalPart cp in criticalParts)
    {
        car& Car1 = getCarByColor(cp.car1_color);
        car& Car2 = getCarByColor(cp.car2_color);

        if (Car1.getColor() != CarColor::WrongColor && Car2.getColor() != CarColor::WrongColor)
        {
            /*
                Critical Part: O-----O
                              CS1   CS2
                Check 4 different situation:
                   - Car1 is in CS1 and Car2 is between CS1 and CS2
                   - Car1 is in CS2 and Car2 is between CS1 and CS2
                   - Car2 is in CS1 and Car1 is between CS1 and CS2
                   - Car2 is in CS2 and Car1 is between CS1 and CS2

            */

            //if Car1 in Pos1
            checkOnCarInOneCriticalPosition(Car1, Car2, cp.car1_pos1, cp.car2_pos1, cp.car2_pos2, cp.car1_criticalDir1, cp.car2_criticalDir2);
            //if Car1 in Pos2
            checkOnCarInOneCriticalPosition(Car1, Car2, cp.car1_pos2, cp.car2_pos2, cp.car2_pos1, cp.car1_criticalDir2, cp.car2_criticalDir1);
            //if Car2 in Pos1
            checkOnCarInOneCriticalPosition(Car2, Car1, cp.car2_pos1, cp.car1_pos1, cp.car1_pos2, cp.car2_criticalDir1, cp.car1_criticalDir2);
            //if Car2 in Pos2
            checkOnCarInOneCriticalPosition(Car2, Car1, cp.car2_pos2, cp.car1_pos2, cp.car1_pos1, cp.car2_criticalDir2, cp.car1_criticalDir1);
        }
    }
}

void scheduler::carStep(int currentTime)
{
    for (std::list<car>::iterator it = cars.begin(); it != cars.end(); ++it)
    {
        it->step(currentTime);
    }
}

void scheduler::checkCollisionPoint(car car1, int pos1, int time, collisionPoint& cp)
{
    if (car1.getColor() != CarColor::WrongColor && car1.getCurrentPos() == pos1)
    {
        arrive arrivedCar;
        arrivedCar.color = car1.getColor();
        arrivedCar.time = time;
        cp.arrives.push_back(arrivedCar);
    }
}

bool scheduler::checkCollisionPoint(car car1, int pos1, car car2, int pos2)
{
    if (car1.getColor() == CarColor::WrongColor || car2.getColor() == CarColor::WrongColor) return false;

    if (car1.getCurrentPos() == pos1 &&
        car2.getCurrentPos() == pos2 &&
        !(car1.getWaiting()) &&
        !(car2.getWaiting()))
    {
        return true;
    }

    return false;
}

void scheduler::updateArrives(int time)
{
    for (std::list<collisionPoint>::iterator cp = collisionPoints.begin(); cp != collisionPoints.end(); ++cp)
    {
        checkCollisionPoint(getCarByColor(cp->color1), cp->pos1, time, *cp);
        checkCollisionPoint(getCarByColor(cp->color2), cp->pos2, time, *cp);
    }
}

void scheduler::checkCollision()
{
    for each (collisionPoint cp in collisionPoints)
    {
        if (checkCollisionPoint(getCarByColor(cp.color1), cp.pos1, getCarByColor(cp.color2), cp.pos2))
        {
            isWorking = false;
        }
    }
}

void scheduler::writeResults()
{
    std::cout << "A kisautok helye: ";
    for each (car carobject in cars)
    {
        std::cout << colorString(carobject.getColor()) << ": " << carobject.getCurrentPos() << " ";
    }
    std::cout << std::endl;

    std::cout << "A kisautok sebessege: ";
    for each (car carobject in cars)
    {
        std::cout << colorString(carobject.getColor()) << ": " << carobject.getVelocity() << " ";
    }
    std::cout << std::endl;

    std::cout << "A kisautok altal megtett teljes korok szama: ";
    for each (car carobject in cars)
    {
        std::cout << colorString(carobject.getColor()) << ": " << carobject.getLaps() << " ";
    }
    std::cout << std::endl;

    std::cout << "A varakozasi idok: ";
    for each (car carobject in cars)
    {
        std::cout << colorString(carobject.getColor()) << ": " << carobject.getWaitingTime() << " ";
    }
    std::cout << std::endl;

    std::cout << "Elso kor megtetelehez szukseges ido: ";
    for each (car carobject in cars)
    {
        std::cout << colorString(carobject.getColor()) << ": " << carobject.getFirstLapTime() << " ";
    }
    std::cout << std::endl;

    std::ofstream outfile("cspont.txt");
    
    for each (collisionPoint cp in collisionPoints)
    {
        outfile << "cs" << cp.id << " ";
        if (cp.arrives.empty()) outfile << "Nem volt athaladas!";
        else
        {
            for each (arrive a in cp.arrives)
            {
                outfile << colorString(a.color) << "(" << a.time << ") ";
            }
        }
        outfile << std::endl;
    }
}

void scheduler::sumEnergy()
{
    energy = 0;
    for each (car carobject in cars)
    {
        energy += carobject.getEnergy();
    }
}

void scheduler::writeEnergyResult()
{
    std::cout << "A 10 perc alatt elhasznlt energia: " << energy << std::endl;
}